/*
 * bloom_test.c:
 *
 * A program to test the nsrl bloom library.
 *
 * The software provided here is released by the National Institute of
 * Standards and Technology (NIST), an agency of the U.S. Department
 * of Commerce, Gaithersburg MD 20899, USA.  The software bears no
 * warranty, either expressed or implied. NIST does not assume legal
 * liability nor responsibility for a User's use of the software or
 * the results of such use.
 *
 * Please note that within the United States, copyright protection,
 * under Section 105 of the United States Code, Title 17, is not
 * available for any work of the United States Government and/or for
 * any works created by United States Government employees. User
 * acknowledges that this software contains work which was created by
 * NIST employees and is therefore in the public domain and not
 * subject to copyright.  The User may use, distribute, or incorporate
 * this software provided the User acknowledges this via an explicit
 * acknowledgment of NIST-related contributions to the User's
 * work. User also agrees to acknowledge, via an explicit
 * acknowledgment, that any modifications or alterations have been
 * made to this software before redistribution.
 * --------------------------------------------------------------------
 *
 * Change History:
 * Simson L. Garfinkel - May 1, 2008
 *   - Major rewrite of original NIST code.using new function
 * Douglas White - douglas.white@nist.gov - June 21, 2003
 *   Original implementation in perl.
 */

#include "config.h"

#include "bloom.h"
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <inttypes.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <errno.h>
#include <err.h>

#include <getopt.h>
#include <ctype.h>
const char *progname = "nsrl_bloom";

FILE *load_file=0;
int  load_line  = 0;
#include <sys/stat.h>
#include <signal.h>
int hash_funcs = 4;
int bloom_bits = 32;

void term_info(int sig)
{
    struct stat st;
    fprintf(stderr,"Line %d ",load_line);
    if(fstat(fileno(load_file),&st)==0){
	if(st.st_size){
	    fprintf(stderr,"(%6.2g%% loaded)",ftello(load_file) *100.0 / st.st_size);
	}
    }
    fprintf(stderr,"\n\n");
    
}

void queryfile(nsrl_bloom *b,const char *fname)
{
    char linebuf[256];
    int  nsrl_file = 1;
    FILE *load_file = 0;

    int load_line = 0;
    if(strcmp(fname,"-")==0){
	load_file = stdin;
    }
    else {
	load_file = fopen(fname,"r");
    }
    if(!load_file){
	perror(fname);
	exit(1);
    }
    while(!feof(load_file) && fgets(linebuf,sizeof(linebuf),load_file)){
	unsigned char hashbuf[120];
	int found_bits;
	char *hash_start=0;
	load_line ++;
	linebuf[sizeof(linebuf)-1] = 0;	/* good hygine */
	nsrl_file = 1;//copied function from nsrl_load
	if(nsrl_file){
	    if(linebuf[0]=='"'){hash_start=&linebuf[1];}
	    else{hash_start=linebuf;} 
	    if(b->hash_bits==128){
		hash_start[32]='\000';	/* terminate */
	    }
	    else {
		hash_start[40]='\000';
	    }
	}

	if(!isxdigit(*hash_start)){
	    fprintf(stderr,"invalid hash: %s\n",hash_start);
	    continue; /* line doesn't begin with hash */
	}
	found_bits = nsrl_hex2bin(hashbuf,sizeof(hashbuf),hash_start);
	if(found_bits != b->hash_bits){
	    char *c = index(linebuf,'\n');
	    if(c) *c = '\000';
	    fprintf(stderr,"Input format error.\n");
	    fprintf(stderr,"%s line %d: '%s' has %d bits (%d expected)\n",
		    fname,load_line,linebuf,found_bits,b->hash_bits);
	    exit(1);
	}
	if(nsrl_bloom_query(b,hashbuf)){
	    printf("%s present\n",hash_start);
	}
	else {
	    printf("%s absent\n",hash_start);
	}
    }
#ifdef SIGINFO
    signal(SIGINFO,0);
#endif
}


void usage()
{
    printf("Usage: %s [options] bloomfile [HASH ...]\n",progname);
    printf("     bloomfile Bloom filter file to be updated\n");
    printf("     -b X         Create Bloom filter of size 2^X bits.\n");
    printf("                  Default is %d.\n",bloom_bits);
    printf("     -c 128       Create the Bloom file for 128-bit hashes (MD5s)\n");
    printf("     -c 160       Create the Bloom file for 160-bit hashes (SHA1s)\n");
    printf("     -i hashfile  Bulk insert codes; first 32 chars in each line are the MD5\n");
    printf("                  (Automatically detects RDS files)\n");
    printf("                  Specify '-' for stdin\n");
    printf("     -a           Add the hash codes on the command line\n");
    printf("     -d           debug mode\n");
    printf("     -k int	      number of bloom vectors to use; default is %d.\n",hash_funcs);
    printf("     -v           verbose output\n");
    printf("     -h           this help\n");
    printf("     HASH...      Optional hash codes (MD5s or SHA1s) to query (or add with -a)\n");
    printf("     -q hashfile  Bulk query\n");
    printf("                  Specify '-' for stdin\n");
    printf("     -s           print bloom stats and histogram\n");
    printf("     -w walkfile  walkfile to test coverage of\n");
    printf("     -p password  specifies password for Bloom filter\n");
}

/**
 * @param b - bloom table to load
 * @param fname - file to load into
 */

void rds_loadfile(nsrl_bloom *b,const char *fname)
{
    if(b->debug) printf("In rds_loadfile.\n");
    char linebuf[256];
    int  total = 0;
    int  nsrl_file = 0;
    FILE *load_file=0;
#ifdef SIGINFO
    signal(SIGINFO,term_info);
#endif
    int load_line = 0;

    load_line = 0;
    if(strcmp(fname,"-")==0){
	load_file = stdin;
    }
    else {
	load_file = fopen(fname,"r");
    }
    if(!load_file){
	perror(fname);
	exit(1);
    }
    while(!feof(load_file) && fgets(linebuf,sizeof(linebuf),load_file)){
	unsigned char hashbuf[120];
	int found_bits;
	char *hash_start=linebuf;
	
	load_line ++;
	linebuf[sizeof(linebuf)-1] = 0;	/* good hygine */

	if(load_line==1){
	    /* See if this is an NSRL file */
	    if(strncmp(linebuf,"\"SHA-1\",\"MD5\"",13)==0){
		puts("This is an NSRL file\n");
		nsrl_file = 1;
		continue;
	    }
	}

	if(nsrl_file){
	    /* Make sure it really is an NSRL-formatted file and, if so,
	     * set the start point properly.
	     */
	    if(linebuf[0]!='"' || linebuf[41]!='"' || linebuf[42]!=',' ||
	       linebuf[43]!='"' || linebuf[76]!='"'){
		fprintf(stderr,"Invalid line %d in RDS file:\n%s\n",load_line,linebuf);
		//fclose(load_file);
		//return;
		//break;
		continue;
	    }
	    if(b->hash_bits==128){
		hash_start=linebuf+44;
		hash_start[32]='\000';	/* terminate */
	    }
	    else {
		hash_start=linebuf+1;
		hash_start[40]='\000';
	    }
	}
	if(b->debug)printf("after if(nsrl_file)\n");
	if(b->debug)printf("%s\n", hash_start);
	if(*hash_start == '"') hash_start++;
	if(b->hash_bits==160)hash_start[40]='\0';
	else hash_start[32]='\0'; 
	if(!isxdigit(*hash_start)){
	    fprintf(stderr,"invalid hash: %s\n",hash_start);
	    continue; /* line doesn't begin with hash */
	}
	found_bits = nsrl_hex2bin(hashbuf,sizeof(hashbuf),hash_start);
	if(found_bits != b->hash_bits){
	    char *c = index(linebuf,'\n');
	    if(c) *c = '\000';
	    fprintf(stderr,"Input format error.\n");
	    fprintf(stderr,"%s line %d: '%s' has %d bits (%d expected)\n",
		    fname,load_line,linebuf,found_bits,b->hash_bits);
	    //exit(1);
	    continue;
	}
	if(b->debug) printf("adding hash %s\n",hash_start);
	nsrl_bloom_add(b,hashbuf);
	total++;
	if(b->debug && total%1000==0){
	    printf("inserted %d; total in bloom filter: %"PRId64"\n",total,b->added_items);
	}
    }
#ifdef SIGINFO
    signal(SIGINFO,0);
#endif
}

void walkfile(nsrl_bloom *b, const char *fname){
    FILE *walk_file=0;
    char linebuf[256];
    char shatoquery[41];
    unsigned char hashbuf[120];
    unsigned int	count = 0, found =0;
    shatoquery[40] = '\0';
    walk_file = fopen(fname, "r");
    if(!walk_file){
	perror(fname);
	exit(1);
    }
	
    while(!feof(walk_file) && fgets(linebuf, sizeof(linebuf), walk_file)){
	if(!strncmp(linebuf, "SHA1: ", 6)){//valid sha1 line
	    strncpy(shatoquery, &linebuf[6], 40);
	    nsrl_hex2bin(hashbuf, sizeof(hashbuf),shatoquery); //convert to binary for query	
	    count++;
	    if(nsrl_bloom_query(b, hashbuf)){
		//			printf("%s present\n", shatoquery);
		found++;
	    }
	    //		else{ printf("%s absent\n", shatoquery);}
	} //end if loop for valid sha1
    }//end while !feof loop
    fclose(walk_file);
    printf("%u found, %u searched.  %f %%coverage.\n", found, count, (((float)found/(float)count)*100));
}


int main(int argc,char **argv)
{
    nsrl_bloom *b;
    const char *file=0;
    int ch=0;
    const char *insert_file=0;
    const char *query_file=0;
    int opt_add = 0;
    int opt_create = 0;
    int opt_debug = 0;
    int histo = 0;
    progname = argv[0];
    const char *walk_file=0;
    int mode = O_RDONLY;

    while ((ch = getopt(argc,argv,"ab:dc:k:i:vw:hsq:")) != -1){
	switch(ch){
	case 'a':
	    opt_add = 1;
	    mode = O_RDWR;
	    break;
	case 'b': bloom_bits = atoi(optarg);  break; 
	case 'c':
	    switch(atoi(optarg)){
	    case 128:
	    case 160:
		opt_create = atoi(optarg);
		break;
	    default:
		fprintf(stderr,"-c option requires 128 or 160 as an argument.\n");
		usage();
		exit(1);
	    }
	    break;
	case 'k': hash_funcs = atoi(optarg); break;
	case 'i':
	    insert_file = optarg;
	    mode = O_RDWR;
	    break;
	case 'd':
	    opt_debug++;
	    break;
	case 's': histo = 1; break;
	case 'q': query_file = optarg; break; 
	case 'w': walk_file = optarg; break;
	default:
	case 'h': usage(); exit(0);
	}//end switch
    }//end while
    argc -= optind;
    argv += optind;

    if(argc<1) usage();
    argc--;
    file = *argv++;

    b = nsrl_bloom_alloc();
    b->debug = opt_debug;
    if(opt_create){
	if(!access(file,R_OK)){
	    perror(file);
	    fprintf(stderr,"Do not supply -c option to use an existing Bloom file.\n");
	    exit(1);
	}
	if(nsrl_bloom_create(b,file,opt_create,bloom_bits,hash_funcs,"NSRL bloom filter")){
	    errx(1,"Cannot create bloom filter?");
	}
    }

    if(!opt_create){
	if(access(file,R_OK)){
	    perror(file);
	    fprintf(stderr,"Use the -c option to create a new Bloom file.\n");
	    exit(1);
	}
	if(nsrl_bloom_open(b,file,mode)){
	    err(1,"nsrl_bloom_open(%s)",file);
	}
	bloom_bits = b->M;
	if(b->debug){
	    char ibuf[4096];
	    nsrl_bloom_info(ibuf,sizeof(ibuf),b);
	    fputs(ibuf,stdout);
	}
    }

    if(insert_file){
	rds_loadfile(b,insert_file);    /* Load hashes from provided insert_file */
    }
    
    if(query_file){
	queryfile(b,query_file);
    }
	
    if(walk_file){
	walkfile(b, walk_file);
    }

    if(histo){
	unsigned int counts[256];
	memset(counts, 0, sizeof(counts));
	nsrl_calc_histogram(b,counts);
	nsrl_print_histogram(b,counts);
    }

    /* Now test hashes that are provided on command line */
    while(*argv){
	char *code = *argv++;
	argc--;
	int bits;
	unsigned char hashbuf[20];
	if(!isxdigit(*code)){
	    fprintf(stderr,"%s is not a hash\n",code);
	    exit(1);
	}
	bits = nsrl_hex2bin(hashbuf,sizeof(hashbuf),code);
	if(strlen(code)!=bits/4) bits=strlen(code)*4;
	if(bits != b->hash_bits){
	    fprintf(stderr,"%s is a %d-bit hash, but %s is a %d-bit bloom filter.\n",
		    code,bits,file,b->hash_bits);
	    exit(1);
	}
	if(opt_add){
	    nsrl_bloom_add(b,hashbuf);
	    printf("%s added\n",code);
	}
	else{
	    if(nsrl_bloom_query(b,hashbuf)){
		printf("%s present\n",code);
	    }
	    else {
		printf("%s absent\n",code);
	    }
	}
    }
    return 0;
}




