/*
 * bloomcat: A program to concatenate multiple bloom files together
 */

#include "config.h"

#include "bloom.h"

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <errno.h>
#include <err.h>

#include <getopt.h>
#include <ctype.h>
#include <sys/stat.h>

#ifdef HAVE_COPYFILE_H
#include <copyfile.h>
#endif

const char *progname = "bloomcat";

void usage()
{
    printf("Usage: %s [options] bloomfile1  [bf2 bf3...]\n",progname);
    printf("Take the contents from one or more bloom files and cat them together\n");
    printf("     -o outfile   output file\n");
    exit(0);
}

int main(int argc,char **argv)
{
    const char *outfile = 0;
    progname = argv[0];
    int ch;
    
    while ((ch = getopt(argc,argv,"o:h?")) != -1){
	switch(ch){
	case 'o': outfile = optarg; break;
	default:
	case '?':
	case 'h': usage(); exit(0);
	}
    }
    argc -= optind;
    argv += optind;

    if(argc<1) usage();
    if(!outfile){
	fprintf(stderr,"No outfile specified!\n");
	usage();
    }

    /* Get the parameters for the first bloom file and compare it to the others, make
     * sure they are all comsurate.
     */

    nsrl_bloom *b1 = nsrl_bloom_alloc();
    if(nsrl_bloom_open(b1,argv[0],O_RDONLY)){
	err(1,"%s",argv[0]);
    }
    
    /* Check all of the other blooms... */
    for(int i=1;i<argc;i++){
	nsrl_bloom *bn = nsrl_bloom_alloc();
	if(nsrl_bloom_open(bn,argv[i],O_RDONLY)){
	    err(1,"%s",argv[i]);
	}
	if(b1->hash_bits != bn->hash_bits) err(1,"%s: hash_bits %d!=%d",argv[i],b1->hash_bits,bn->hash_bits);
	if(b1->M != bn->M) err(1,"%s: bloom_bits %d!=%d",argv[i],b1->M,bn->M);
	if(b1->k != bn->k) err(1,"%s: k %d!=%d",argv[i],b1->k,bn->k);
	if(b1->vector_bytes != bn->vector_bytes){
	    errx(1,"%s: vector_bytes %zd!=%zd",argv[i],b1->vector_bytes,bn->vector_bytes);
	}
	nsrl_bloom_free(bn);
    }

    /* Looks good. Now add them together for real */
    int fd = open(outfile,O_RDWR|O_CREAT|O_EXCL,0666);
    if(fd<0) err(1,"%s",outfile);

    /* Copy the first file! */
    lseek(b1->fd,0,SEEK_SET);
    lseek(fd,0,SEEK_SET);
#ifdef HAVE_COPYFILE_H
    if(fcopyfile(b1->fd,fd,NULL,COPYFILE_ALL)) err(1,"fcopyfile");
#else
    char buf[65536];
    int r;
    while((r = read(b1->fd,buf,sizeof(buf)))>0){
	if(write(fd,buf,r)!=r) err(1,"write");
    }
#endif    

    nsrl_bloom_free(b1);
    close(fd);


    /* Open the copy */
    nsrl_bloom *bc = nsrl_bloom_alloc();
    if(nsrl_bloom_open(bc,outfile,O_RDWR)){
	err(1,"%s",outfile);
    }
    
    /* Do the "or" */
    for(int i=1;i<argc;i++){
	nsrl_bloom *bn = nsrl_bloom_alloc();
	if(nsrl_bloom_open(bn,argv[i],O_RDONLY)){
	    for(u_int j = 0; j<bn->vector_bytes;j++){
		bc->vector[j] |= bn->vector[j];
	    }
	}
	nsrl_bloom_free(bn);
    }
    nsrl_bloom_free(bc);
    exit(0);
}




