/*
 * frag_find: given a file, find all of its fragments on the hard drive.
 * Automatically ignore popular fragments. 
 *
 * Performance Log:
 * Version 1.1.1 - 20724 SHA1 calcuations
 */

#include "config.h"

#define __STDC_FORMAT_MACROS
#include <inttypes.h>

#include "bloom.h"
#include "aftimer.h"
#include "adler32.h"
#include "sha1_t.h"
#include "threaded_hash.h"

#include <assert.h>
#include <getopt.h>
#include <stdio.h>
#include <stdint.h>
#include <signal.h>
#include <string.h>
#include <cstring>
#include <fcntl.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#if defined(HAVE_SYS_DISK_H) 
#include <sys/disk.h>
#endif
#if defined(HAVE_SYS_MOUNT_H)
#include <sys/mount.h>
#endif

#include <map>
#include <vector>
#include <bitset>

#ifndef DEFAULT_BLOCKSIZE
#define DEFAULT_BLOCKSIZE 512
#endif

#include "blockfile.h"
#include "xml.h"

using namespace std;

string _e = string("");

#ifdef HAVE_ERR_H
#include <err.h>
#else
void err(int eval,const char *fmt,...)
{
  va_list ap;
  va_start(ap,fmt);
  vfprintf(stderr,fmt,ap);
  va_end(ap);
  fprintf(stderr,": %s\n",strerror(errno));
  exit(eval);
}
#endif

#ifndef HAVE_ATOI64
int64_t atoi64(const char *buf)
{
    int64_t ret=0;
    sscanf(buf,"%"PRId64,&ret);
    return ret;
}
#endif


blockfile imagefile;			// the file we are reading
u_int blocksize = DEFAULT_BLOCKSIZE;
bool use_prefilter = 1;
int opt_raw = 0;


uint64_t blocknumber=0;
void status(int n)
{
    printf("current block: %"PRId64"\n",blocknumber);
}

/** blocklist is a list of block numbers.
 */

class blocklist_t:public std::vector<int64_t> {
public:
    /** return true if the blocklist contains a given block number */
    bool contains(int64_t block) const {
	for(blocklist_t::const_iterator i = this->begin();
	    i != this->end();
	    i++){
	    if(*i == block) return true;
	}
	return false;
    }
};

/* A filemap is a vector of blocklists.
 * The key is the block number of the mapped file, so a file that is 64k in length
 * would have 128 entries. Each element is a blocklist of all the locations
 * on the disk image where the block appears.
 */
class filemap_t:public std::vector<blocklist_t> {
public:
    /* Clean the filemap. If any sector in the input file is in more than
     * one place in the image, check to see if the previous next is in just one place.
     * If so, then only pick the sector that aligns.
     */
    void clean(){
	for(int64_t b=1;b<(int64_t)this->size();b++){
	    if((*this)[b-1].size()==1 && (*this)[b].size()>1){
		int64_t predicted = (*this)[b-1].front()+1;
		if((*this)[b].contains(predicted)){
		    (*this)[b].clear();
		    (*this)[b].push_back(predicted);
		}
	    }
	}
	
	/* Clean in the reverse direction */
	for(int64_t b=this->size()-2;b>=0;b--){
	    if((*this)[b+1].size()==1 && (*this)[b].size()>1){
		int64_t predicted = (*this)[b+1].front()-1;
		if((*this)[b].contains(predicted)){
		    (*this)[b].clear();
		    (*this)[b].push_back(predicted);
		}
	    }
	}
    }

    /* Return the length of a given mapping run --- the number of sequential blocks
     * starting at a given location.
     */
    int64_t runlen(int64_t target_block,int64_t image_block){
	int64_t len = 1;
	for(len = 1;target_block + len < (int64_t)this->size(); len++){
	    if(! (*this)[target_block+len].contains(image_block+len)){
		return len;
	    }
	}
	return len;
    }

    int64_t found_count(){
	int64_t count = 0;
	for(int64_t b=0;b<(int64_t)this->size();b++){
	    if ((*this)[b].size()>0) count++;
	}
	return count;
    }

    double found_percent(){
	return found_count() * 100.0 / this->size();
    }
};



/** sha1map maps individual sha1's to a list of blocks.
 * It has to be a lst of blocks, because multiple blocks may have the same SHA1.
 */

class sha1map_t:public std::map<class sha1_t,blocklist_t> {
};

/** The prefilter */

class prefilter_t:public std::bitset<16777216> {
    /** return a very fast-to-calculate 24-bit value from a 512-byte buffer*/
    static uint32_t val3char(const u_char *buf) {
	return av_adler32_update(0,buf,512) & 0xffffff;
    }
public:
    int64_t hits;
    prefilter_t():hits(0){
    }
    double utilization() {
	int64_t seen = 0;
	for(int i=0;i<16777216;i++){
	    if((*this)[i]) seen++;
	}
	return (double)seen / 16777216.0;
    }
    void set_from_buf(const u_char *buf){
	(*this)[val3char(buf)] = 1;
    }
    bool query_from_buf(const u_char *buf){
	bool ret = (*this)[val3char(buf)] == 1;
	if(ret) hits++;
	return ret;
    }
};
prefilter_t prefilter;	// bitset to hold first 3 bytes of block


int opt_M = 32;
void usage()
{
    printf("usage: frag_find [options] <raw disk file> <target1> [<target2...>] \n");
    printf("\n");
    printf("options:\n");
    printf("     -s <start>   - start image scan at block <start> (default is beginning)\n");
    printf("     -e <end>     - end image scan at block <start> (default is end)\n");
    printf("     -b blocksize - sets the blocksize (default is %d)\n",DEFAULT_BLOCKSIZE);
    printf("     -r           - prints the raw association map, in addition to the cleaned on\n");
    printf("     -S           - print stats\n");
    printf("     -MN	      - specifies size of bloom filter, in 2^N bits (default %d)\n",opt_M);
    printf("     -x1          - do not use bloom filter or prefilter\n");
    printf("     -x2          - do not use prefilter\n");
    printf("     -Xfname.xml  - output digital forensics XML file to fname.xml\n");
    exit(1);
}

/* A targetfile is like a blockfile, but it also keeps track of:
 * -  a sha1map, which is the block numbers in the target where each sha1 exists.
 * - a filemap, an array that holds all of the block lists
 */
class targetfile:public blockfile {
public:;
    sha1map_t sha1map;
    filemap_t filemap;
    void load(const char *fname,NSRLBloom &b);
    char sha1[41];
    void print_report(class xml *);
};

void targetfile::print_report(class xml *x) {
    printf("Target Block(s)     Found at image block\n");
    for(uint64_t tb=0;tb<(uint64_t)filemap.size();){
	if (filemap[tb].size()==0){
	    tb++;			// target block not in image
	    continue;
	}
	uint64_t first_ib = filemap[tb].front();
	uint64_t run_blocks = filemap.runlen(tb,first_ib);
	uint64_t run_bytes = run_blocks * blocksize;

	/* If run_bytes extends past the end of the file, shorten it to the file length */
	if(tb*blocksize + run_bytes > filesize){
	    run_bytes = filesize - tb*blocksize;
	}

	if(run_blocks>1){
	    printf("%10"PRId64"-%-10"PRId64"    %10"PRId64"-%-10"PRId64"   (%"PRId64" blocks)\n",
		   tb,tb+run_blocks-1, first_ib, first_ib+run_blocks-1, run_blocks);
	    if(x){
		char buf[1024];
		snprintf(buf,sizeof(buf),"file_offset='%"PRId64"' img_offset='%"PRId64"' len='%"PRId64"'",
			 tb*blocksize,first_ib*blocksize,run_bytes);
		x->xmlout(string("run"),_e,string(buf),false);
	    }
	    tb += run_blocks;
	    continue;
	}
	bool first=true;
	for(blocklist_t::const_iterator i = filemap[tb].begin();
	    i != filemap[tb].end();
	    i++){
	    if(first){
		printf("%10"PRId64"               ",tb);
		first = false;
	    }
	    printf(" %10"PRId64" ",*i);
	    if(x){
		char buf[1024];
		snprintf(buf,sizeof(buf),"file_offset='%"PRId64"' img_offset='%"PRId64"' len='%"PRIu64"'",
			 tb*blocksize,(*i) * blocksize,run_bytes);
		x->xmlout(string("run"),_e,string(buf),false);
	    }
	}
	if(first==false) printf("\n");
	tb++;
    }

    /* Now see if the partial block at the end of any of the possible runs match */
    int extra_bytes = filesize % blocksize;
    if(extra_bytes){
	for(blocklist_t::const_iterator i=filemap.back().begin();
	    i!=filemap.back().end();
	    i++){

	    bool done = false;
	    uint64_t last_image_block  = *i;
	    uint64_t last_target_block = filemap.size()-1;

	    /* See if the next block has the data... */
	    u_char *image_extra = (u_char *)calloc(extra_bytes,1);
	    u_char *target_extra = (u_char *)calloc(extra_bytes,1);
	    
	    ssize_t ilen = imagefile.pread((void *)image_extra,extra_bytes,(last_image_block+1)*blocksize);
	    ssize_t tlen = this->pread((void *)target_extra,extra_bytes,(last_target_block+1)*blocksize);
	    if(tlen==extra_bytes && ilen==extra_bytes && memcmp(target_extra,image_extra,extra_bytes)==0){
		if(x){
		    char buf[1024];
		    snprintf(buf,sizeof(buf),"file_offset='%"PRId64"' img_offset='%"PRId64"' len='%d'",
			     (last_target_block+1)*blocksize,
			     (last_image_block+1)*blocksize,
			     extra_bytes);
		    x->xmlout(string("run"),_e,string(buf),false);
		}
		done=true;
	    }
	    free(image_extra);
	    free(target_extra);
	    if(done) break;
	}
    }
}


void targetfile::load(const char *fname,NSRLBloom &b)
{
    if(this->open(fname,blocksize)<0){
	err(1,"Error: Cannot open %s",fname);
    }
	
    if(imagefile.filesize < this->filesize ){
	err(1,"Error: Target file %s is larger than image file; cannot continue\n",
	    fname);
    }
	    
    printf("Blocks: %"PRId64" (%d-byte blocks)\n", this->blocks,blocksize);
    sha1_t *obj = sha1_t::new_from_file(fname);
    if(!obj) err(1,"cannot compute sha1 of %s",fname);
    obj->hex(sha1);
    delete obj;
    printf("SHA1:   %s\n",sha1);

    /** Read through the target file block by block and create a SHA1 structure
     * for each block that is read. Each SHA1 must be allocated because we need a place
     * to keep all of these SHA1s. The map points to each SHA1.
     */
    for(blocknumber=0;blocknumber < this->blocks; blocknumber++){
	const u_char *buf = this->getblock(blocknumber);
	if(use_prefilter) prefilter.set_from_buf(buf);
	sha1_t *sha1 = new sha1_t(buf,blocksize);

	// report if it is in the input file twice
	if(opt_raw && b.query(sha1->sha1)){
	    char hexbuf[41];
	    printf("Input file block %"PRId64" (%s) appears previously at ",blocknumber,sha1->hex(hexbuf));
	    for(blocklist_t::const_iterator i = this->sha1map[*sha1].begin();
		i!=this->sha1map[*sha1].end();
		i++){
		printf("%"PRId64" ",*i);
	    }
	    printf("\n");
	}

	// add to the map and to the bloom filter
	this->sha1map[*sha1].push_back(blocknumber);
	b.add(sha1->sha1);
    }
    this->filemap.resize(this->blocks);
}

class targets_t:public std::vector<targetfile *>{
};


int main(int argc,char **argv)
{
    targets_t targets;
    uint64_t opt_start = 0;
    uint64_t opt_end   = 0;
    int opt_stats = 0;
    int ch;
    bool use_bloom = 1;
    uint64_t bloom_false_positives=0;
    class xml *x = 0;
    string command_line;

    /* Make a copy of the command line */
    for(int i=0;i<argc;i++){
	if(i>0) command_line.push_back(' ');
	command_line.append(argv[i]);
    }

    while ((ch = getopt(argc,argv,"s:e:rb:Sh?x:X:M:")) != -1){
	switch(ch){
	case 's': opt_start = atoi64(optarg);break;
	case 'e': opt_end   = atoi64(optarg);break;
	case 'r': opt_raw   = 1;break;
	case 'b': blocksize = atoi(optarg); break;
	case 'S': opt_stats++;break;
	case 'M': opt_M     = atoi(optarg); break;
	case 'X':
	    switch(atoi(optarg)){
	    case 1: use_bloom = 0;break;
	    case 2: use_prefilter = 0;break;
		break;
	    }
	    break;
	case 'x':
	    x = new xml();
	    x->set_outfilename(optarg);
	    x->set_makeDTD(true);
	    break;
	case 'h':
	case '?':
	default:
	    usage();
	}
    }
    if(opt_M <5 || opt_M > 32){
	fprintf(stderr,"M must be between 5 and 32.\n");
	exit(1);
    }

    argc -= optind; 
    argv += optind;
    
    if(argc<2) usage();

    if(blocksize<4){
	printf("Blocksizes smaller than 4 are meaningless. Try again.\n");
	exit(1);
    }

    printf("Creating bloom filter...\n");
    NSRLBloom b;
    if(b.create(0,160,opt_M,5,"In-memory bloom filter for frag find")){
	printf(">>> not enough memory for bloom filter\n");
	exit(1);
    }
    else {
	printf("Bloom filter created.\n");
    }

#ifdef SIGINFO
    signal(SIGINFO,status);
#endif

    if(imagefile.open(*argv,blocksize)<0) err(1,"Cannot open %s",imagefile.filename);
    if(imagefile.filesize==0){
	err(1,"Error: Image size %s is zero.\n",*argv);
    }

    if(imagefile.filesize < blocksize){
	err(1,"Image size (%"PRId64") is smaller than blocksize (%d); cannot continue.\n",
	    imagefile.filesize,blocksize);
    }

    argv++;

    /* Load up all of the target files */
    for(;*argv;argv++){
	printf("\n");
	printf("Search file: %s (file #%d)\n",*argv,(int)targets.size()+1);
	targetfile *t = new targetfile();
	t->load(*argv,b);
	targets.push_back(t);
    }

    printf("Now searching image file...\n");

    uint64_t count = 0;

    /** Read through the image file block by block. This time we can use a shared SHA1 structure
     * because we do not need to allocate a hash for each block.
     */

    if(opt_end==0) opt_end = imagefile.blocks;
    aftimer timer;
    timer.start();
    for(uint64_t blocknumber=opt_start;blocknumber < opt_end && blocknumber < imagefile.blocks; blocknumber++){

	/* If this is one of the 100,000 even blocks, print status info */
	if(blocknumber>opt_start && (((blocknumber-opt_start) % 100000)==0)){
	    int64_t blocks = blocknumber-opt_start;
	    int64_t total = opt_end-opt_start;
	    putchar('\r');
	    printf("%"PRId64"M out of %"PRId64"M sectors processed; hits=%"PRId64,
		   blocks/1000000,total/1000000,count);
	    if(timer.elapsed_seconds()>1.0){

		int tseconds = timer.eta((double)blocks/(double)total);
		int tminutes =  tseconds / 60;
		int days = tseconds / (24*60*60);
		int hours = (tminutes / 60) % 24;
		int minutes = tminutes % 60;
		int seconds = tseconds % 60;

		printf("; %3g Kblocks/sec; done in ", (blocks/1000)/timer.elapsed_seconds());
		if(days) printf("%d days ",days);
		printf("%d:%02d:%02d ",hours,minutes,seconds);
	    }
	    fflush(stdout);
	}

	/* Scan through the input file block-by-block*/
	const u_char *buf = imagefile.getblock(blocknumber);

	/* If block is not in prefilter, then there is no sense to compute hash */
	if(use_prefilter){
	    if(prefilter.query_from_buf(buf)==0){
		continue;
	    }
	}
	/* Compute the SHA1 */
	struct sha1_t sha1(buf,blocksize);

	/* If the block's SHA1 is not in the bloom filter, there is no need to scan for it in the file maps. */

	if(b.query(sha1.sha1)==0) continue;

	/* Get the list of blocks where that SHA1 was observed and add each one to the filemap.
	 */
	if(opt_raw && use_bloom) printf("Block %"PRId64" SHA1 in bloom filter.\n",blocknumber);
	if(opt_raw) printf("   >>> found at ");
	int64_t added = 0;
	for(targets_t::iterator t = targets.begin(); t!=targets.end(); t++){
	    for(blocklist_t::const_iterator i = (*t)->sha1map[sha1].begin();
		i!=(*t)->sha1map[sha1].end();
		i++){
		if(opt_raw) printf("   %"PRId64" ",*i);
		(*t)->filemap[*i].push_back(blocknumber);
		added++;
	    }
	}
	if(opt_raw) printf("\n");
	if(added==0) bloom_false_positives++;
	count++;
    }


    printf("\n\n");
    printf("Image file:  %s  (%"PRId64" blocks)\n",imagefile.filename,imagefile.blocks);
    printf("Blocksize: %d\n",targets[0]->blocksize);

    /* If we got this far, it's time to make the XML file */
    if(x){
	x->open();
	x->push("frag_find","xmloutputversion='0.3'");
	x->push("metadata",
		"\n  xmlns='http://afflib.org/frag_find/' "
		"\n  xmlns:xsi='http://www.w3.org/2001/XMLSchema-instance' "
		"\n  xmlns:dc='http://purl.org/dc/elements/1.1/'" );
	x->xmlout("dc:type","Hash-Based Carving Report","",false);
	x->pop();

	/* Output carver information per photorec standard */
	x->push("creator");
	x->xmlout("program","frag_find");
	x->xmlout("version",PACKAGE_VERSION);
	x->add_DFXML_build_environment();
	x->add_DFXML_execution_environment(command_line);
	x->pop();			// creator
	x->push("source");
	x->xmlout("image_filename",imagefile.filename);
	x->xmlprintf("blocks","","%d",imagefile.blocks);
	x->xmlprintf("blocksize","","%d",targets[0]->blocksize);
    }


    for(targets_t::iterator t = targets.begin(); t!=targets.end(); t++){
	printf("-------------------\n");
	printf("Target file: %s  (%"PRId64" blocks)\n",(*t)->filename,(*t)->blocks);
	if(x){
	    x->push("fileobject");
	    x->xmlout("filename",(*t)->filename);
	    x->xmlprintf("blocks",_e,"%d",(*t)->blocks);
	    x->xmlprintf("filesize",_e,"%"PRId64,(*t)->filesize);
	    x->xmlout("hashdigest",(*t)->sha1,string("type='sha1'"),false);
	}
	if(opt_start!=0){
	    printf("   NOTE: Image scan started at block %"PRId64"\n",opt_start);
	}
	if(opt_end!=imagefile.blocks){
	    printf("   NOTE: Image scan ended at block %"PRId64"\n",opt_end);
	}

	printf("Blocks of target file found in image file: %"PRId64"\n",count);
	printf("Here is where they were found:\n");

	if(opt_raw){
	    printf("RAW FILEMAP:\n");
	    printf("===============================================================================\n");
	    printf("\n");
	    (*t)->print_report(0);
	}

	(*t)->filemap.clean();

	/* note: perhaps it would make more sense to iteratively clean. */
	(*t)->print_report(x);
	printf("Total blocks of original file found: %"PRId64" (%2.0f%%)\n",
	       (*t)->filemap.found_count(),
	       (*t)->filemap.found_percent());
	if(opt_stats){
	    double runtime = timer.elapsed_seconds();
	    printf("Total number of prefilter hits: %"PRId64" (%"PRId64" blocks filtered out)\n",
		   prefilter.hits,(imagefile.blocks-opt_start)-prefilter.hits);
	    printf("Total number of bloom filter hits: %"PRId64" (%"PRId64" blocks filtered out)\n",
		   b.hits,(imagefile.blocks-opt_start)-b.hits);
	    printf("Total number of bloom false positives: %"PRId64" (should be low)\n",bloom_false_positives);
	    printf("Total number of SHA1 codes calculated: %"PRId64" (should be low)\n",sha1_ctr);
	    printf("Prefilter Utilization: %5.2f%% (should be high)\n",prefilter.utilization()*100.0);

	    printf("Bloom filter utilization: %5.2f%% (ideally less than 50%%)\n",b.utilization()*100.0);
	    printf("Time to run: %4.2g seconds\n",runtime);
	}
	if(x) x->pop();
	}

    if(x){
	x->pop();
	x->pop();
	x->close();
	delete x;
    }

    return 0;
}
