#define __STDC_FORMAT_MACROS
#include <inttypes.h>


#include "config.h"
#include "bloom.h"
#include "sha1_t.h"

#include <sys/types.h>
#include <sys/resource.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <err.h>

using namespace std;
#include <vector>
#include <iostream>
#include <map>
#include <set>
#include <getopt.h>
#include <ctype.h>

/**
 * The counting bloom is like a regular bloom except it also keeps track
 * of how many of each object it founds.
 *
 * Sample line:
 * INSERT INTO blockhash (filekey, hashval, hashtype, blocksize, block) VALUES ('9AF318B4749D308A6366A18DF0A7802D83E50EEF', '166092F87CAE2625B2A735A4039709778C54F085', 'SHA1B4096', 4096, 48);
 */

class sha1_t *justhash = 0;
class sha1_tally:public std::multimap<uint64_t,sha1_t> {};

class sha1set:public std::set<class sha1_t> {};

class CountingBloom : public NSRLBloom {
public:
    class sha1counter_t:public std::map<class sha1_t,uint64_t> { };
    sha1counter_t counter;
    void add(const sha1_t &sha1){
	if(NSRLBloom::query(sha1.sha1)){
	    if(counter[sha1]==0) counter[sha1]=1;
	    counter[sha1]+=1;
	}
	else {
	    NSRLBloom::add(sha1.sha1);
	}
    }
};

const char *progname = "sector_hash";
int opt_x = 0;
bool opt_dups = true;

/** Process the NSRL SQL block insert statement.
 * (Not the most efficient way to distribute block hashes)
 */

uint64_t hashcount=0;
uint64_t filecount=0;

void print_stats(const string &fname) 
{
    printf("%s filecount:%"PRId64"  hashcount:%"PRId64"\n",fname.c_str(),filecount,hashcount);
}

uint64_t freq=1000*1000;

void process(class CountingBloom &b,string fname)
{
    const char *match="VALUES";
    FILE *infile = 0;
    bool pipe = false;
    char buf[4096];

    if(strstr(fname.c_str(),".zip")){
	string cmd = "unzip -p "+fname;
	cout << cmd << "\n";
	infile = popen(cmd.c_str(),"r");
	if(!infile) err(1,"%s",cmd.c_str());
	pipe = true;
    }
    if(!infile){
	infile = fopen(fname.c_str(),"r");
	if(!infile) err(1,"%s",fname.c_str());
    }

    int lastnl=0;
    int lastvalues=0;
    
    std::set<class sha1_t> fileset;	// sha1s seen in the file
    while(fgets(buf,sizeof(buf),infile)){
	if(buf[lastnl]!='\n'){		// assume each line is the same length unless we learn otherwise
	    char *cc = strchr(buf,'\n');
	    if(!cc) break;		// no complete line?
	    lastnl = cc-buf;
	}
	buf[lastnl]=0;
	if(strcmp(buf+lastvalues,match)!=0){
	    const char *cc = strstr(buf,match);
	    if(!cc) continue;		// doesn't have a VALUES line
	    lastvalues = cc-buf;
	}
	const char *data = buf+lastvalues;

	sha1_t filesha1,blocksha1;

	if(filesha1.sethex(data+9)!=160){
	    fprintf(stderr,"invalid line at 1: %s. data+9=%s sethex=%d\n",buf,data+9,filesha1.sethex(data+9));
	    exit(1);
	}
	if(blocksha1.sethex(data+53)!=160){
	    fprintf(stderr,"invalid line at 2: %s; data+53=%s\n",buf,data+53);
	    exit(1);
	}
	int blocknumber = atoi(data+115);
	if(blocknumber==1){
	    filecount++;
	    fileset.clear();
	}

	if(opt_dups){
	    if(fileset.find(blocksha1.sha1) ==fileset.end()){
		fileset.insert(blocksha1.sha1);
		b.add(blocksha1.sha1);
	    }
	}
	if(justhash && *justhash == blocksha1){
	    puts(buf);
	}

	hashcount++;
	if(hashcount % freq == 0) print_stats(fname);
    }
    if(pipe) pclose(infile);
    else fclose(infile);
    print_stats(fname);
}

void print_report(class CountingBloom &b)
{
    int64_t dups = 0;

    /* Now go through all of the dups. Report the most popular SHAs */
    sha1_tally tally;
    for(CountingBloom::sha1counter_t::const_iterator i = b.counter.begin(); i != b.counter.end(); i++){
	char hexbuf[41];
	uint64_t count = (*i).second;
	
	dups += count;
	if(count>100){			// just report those that are in more than 100 locations
	    tally.insert( pair<uint64_t,sha1_t> (count,(*i).first) );
	}
    }
    printf("total hashes encountered: %"PRIu64"d\n",b.added_items);
    printf("Number of block hashes that are duplicated: %"PRIu64" (%f%%)\n",dups,((double)dups)*100.0/b.added_items);
    printf("Bloom filter utilization: %f%%\n",b.utilization()*100.0);;
    uint64_t prev = 0;
    for(sha1_tally::const_iterator i = tally.begin();
	i != tally.end();
	i++){
	char hex[41];
	if((*i).first!=prev){
	    printf("\n%"PRIu64": ",(*i).first);
	    prev = (*i).first;
	}
	printf("%s ",(*i).second.hex(hex));
    }
    printf("\n");
    printf("hash count: %"PRIu64"\n",hashcount);
    printf("file count: %"PRIu64"\n",filecount);
}

void usage()
{
    printf("Usage: %s [options] bloomfile \n", progname);
    printf("     -r file/dir  read the file and report stats\n");
    printf("     -j <hash>    Just report stats for a particular hash\n");
    printf("     -h           Shows this usage.\n");
    printf("     -w <file>    write the bloom file out to the specified file\n");
    printf("\n");
}

int main(int argc,char **argv)
{
    class CountingBloom b;
    vector<string> rfiles;
    const char *outfile = 0;
    int ch;
    while((ch = getopt(argc, argv, "j:r:w:h?")) != -1){
        switch(ch){
	case 'j':
	    opt_dups = false;
	    justhash = sha1_t::new_from_hex(optarg);
	    break;
	case 'r':
	    rfiles.push_back(optarg);
	    break;
	case 'w':
	    outfile = optarg;
	    break;
	default:
	case '?':
	case 'h':
	    usage();
	    exit(0);
        }
    }

    argc -= optind;  
    argv += optind;  

    if(b.create(0,160,34,4,"Standard hash with 2^32 elements")){
	err(1,"%s: Cannot create bloom file: ",outfile);
    }
    for(vector<string>::const_iterator i = rfiles.begin();i!=rfiles.end();i++){
	// If it is a directory, process all of the zip files in the directory
	string fname = *i;
	struct stat st;
	if((stat(fname.c_str(),&st)==0) && S_ISDIR(st.st_mode)){
	    DIR *dirp = opendir(fname.c_str());
	    struct dirent *dp;
	    while ((dp = readdir(dirp)) != NULL){
		if(strstr(dp->d_name,"block_inserts")){
		    process(b,fname + "/" + dp->d_name);
		}
	    }
	    closedir(dirp);
	}
	else {
	    process(b,*i);
	}
    }
    print_report(b);
    nsrl_bloom_print_usage();
    if(outfile){
	cout << "writing " << outfile << "\n";
	if(b.write(outfile)){
	    err(1,"write failed");
	}
    }
    return 0;
}
