#ifndef SHA1_T_H
#define SHA1_T_H

#include <stdint.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>



/* SHA1 class.
 * This class both represents the SHA1 and can calculate it from a supplied buffer.
 *
 * This instrumented sha1 keeps track of how many times it has been called and 
 * has a bypass for calculating SHA1 of all NULLs.
 * 
 * Uses the OpenSSL EVP class
 */

#ifdef HAVE_OPENSSL_HMAC_H
#include <openssl/hmac.h>
#else
#error Requires OpenSSL hmac.h
#endif


static uint64_t sha1_ctr=0;			// number of SHA1 operations performed
static u_char sha1_512nulls[20];
class sha1_t {
public:
    u_char  sha1[20];			// the sha1 value, once computed
    EVP_MD_CTX ctx;			// the context for computing the value
    bool    first;
    sha1_t():first(true){
    }
    /** create a sha1_t object from a sha1 */
    sha1_t(const u_char sha1[20]):first(false){
	memcpy(this->sha1,sha1,20);
    }
    /** Compute a sha1 from a buffer */
    sha1_t(const u_char *buf,size_t bufsize):first(true){
	/* First time through find the SHA1 of 512 NULLs */
	if(sha1_ctr==0){
	    u_char b2[512];
	    EVP_MD_CTX ctx;
	    unsigned int len = sizeof(sha1_512nulls);
	    memset(b2,0,sizeof(b2));
	    EVP_DigestInit(&ctx,EVP_sha1());
	    EVP_DigestUpdate(&ctx,b2,sizeof(b2));
	    EVP_DigestFinal(&ctx,sha1_512nulls,&len);
	    sha1_ctr++;
	}

	/* If the input is 512 bytes long and all NULLs, use our SHA1 of 512 NULLs */
	if(bufsize==512 && iszero(buf,bufsize)){
	    memcpy(sha1,sha1_512nulls,sizeof(sha1));
	    return;
	}
	update(buf,bufsize);
	final();
	sha1_ctr++;
    }
    /** Static method allocator */
    static sha1_t *new_from_hex(const char *hex){
	sha1_t *obj = new sha1_t();
	obj->sethex(hex);
	return obj;
    }

    /** Static method allocateor */
    static sha1_t *new_from_file(const char *fname){
	sha1_t *obj = new sha1_t();
	int fd = open(fname,O_RDONLY);
	if(fd<0) return 0;
	struct stat st;
	if(fstat(fd,&st)<0){
	    close(fd);
	    return 0;
	}
	const u_char *buf = (const u_char *)mmap(0,st.st_size,PROT_READ,MAP_FILE|MAP_SHARED,fd,0);
	if(buf==0){
	    close(fd);
	    return 0;
	}
	obj->update(buf,st.st_size);
	obj->final();
	munmap((void *)buf,st.st_size);
	close(fd);
	return obj;
    }

    /* set the SHA1 to be from a hex string.
     * returns the number of bits that were used.
     */
    int sethex(const char *hex) {
	return nsrl_hex2bin(sha1,sizeof(sha1),hex);
    }

    void update(const u_char *buf,size_t bufsize){
	if(first){
	    EVP_DigestInit(&ctx,EVP_sha1());
	    memset(sha1,0,sizeof(sha1));
	    first=false;
	}
	EVP_DigestUpdate(&ctx,buf,bufsize);
    }
    void final(){
	unsigned int len = sizeof(sha1);
	EVP_DigestFinal(&ctx,sha1,&len);
    }

    /* hexbuf returns the hex representation of the SHA1 */
    const char *hex(char hexbuf[41]) const{
	for(int i=0;i<20;i++){
	    sprintf(hexbuf+i*2,"%02x",sha1[i]);
	}
	return hexbuf;
    }

    /* Static function to determine if something is zero */
    static bool iszero(const u_char *buf,size_t bufsize){
	for(u_int i=0;i<bufsize;i++){
	    if(buf[i]!=0) return false;
	}
	return true;
    }
    /** Operator makes this class work with STL maps.
     */
    bool operator<(const sha1_t &s2) const
    {
	return memcmp(this->sha1,s2.sha1, 20) < 0;
    }
    bool operator==(const sha1_t &s2) const
    {
	return memcmp(this->sha1,s2.sha1, 20) == 0;
    }
};

#endif
